package top.ninwoo.edgecenter.controller;

import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.web.bind.annotation.*;
import top.ninwoo.edgecenter.entity.ClusterConfig;
import top.ninwoo.edgecenter.entity.ContainerInfo;
import top.ninwoo.edgecenter.entity.NetworkTopology;
import top.ninwoo.edgecenter.service.ClusterService;
import top.ninwoo.edgecenter.service.TopologyService;
import top.ninwoo.utils.entity.DockerContainer;
import top.ninwoo.utils.entity.NetworkInfo;
import top.ninwoo.utils.entity.OvsBridge;

import java.util.*;
import java.util.concurrent.*;

@RestController
public class IndexController {
    ConcurrentHashMap<String, ContainerInfo> containerInfoMap = new ConcurrentHashMap<>();
    ConcurrentHashMap<String, Future> scheduledFutureMap = new ConcurrentHashMap<>();
    ScheduledExecutorService scheduledExecutorService = new ScheduledThreadPoolExecutor(8);

    @Autowired
    ClusterService clusterService;

    @Autowired
    TopologyService topologyService;

    @RequestMapping("/index")
    public List<DockerContainer> index(int flag) {

        List<DockerContainer> containerIds = clusterService.getContainerIds(flag == 0);
        return containerIds;
    }

    @RequestMapping("/ovs")
    public List<OvsBridge>  getOvsBridgeList() {
        List<OvsBridge> bridges = clusterService.getOvsBridges();
        return bridges;
    }

    /**
     * 查询容器的网络信息，根据容器的id，获取对应集群的全部节点网络实时情况
     * @param clusterId
     * @param containerId
     * @return
     */
    // 这里一次查询需要两秒，需要采用异步进行查询
    @RequestMapping("/containerMonitor")
    public List<ContainerInfo> getContainerMonitor(@RequestParam("clusterId") Long clusterId,@RequestParam("containerId") String containerId) {
        // 获取全部的容器id
        Set<String> cids = clusterService.getContainerIdsByClusterId(clusterId, containerId);

        // 根据容器id获取全部的
        //List<NetworkInfo> networkTraffic = clusterService.getNetworkTraffic(containerIdsByClusterId);
        List<ContainerInfo> containerInfos = new ArrayList<>();
        cids.forEach( cid -> {
            if(containerInfoMap.containsKey(cid)) {
                ContainerInfo networkInfo = containerInfoMap.get(cid);
                containerInfos.add(networkInfo);
            }
            // TODO: 这里还可以再添加新的处理方式
        });
        return containerInfos;
    }

    @RequestMapping("/enableContainerMonitor")
    public boolean enableNetworkMonitor(long clusterId, String containerName) {
        // 先获取全部的容器id
        Set<String> cids = clusterService.getContainerIdsByClusterId(clusterId, containerName);
        if(cids == null || cids.isEmpty()) {
            return false;
        }
        cids.forEach(cid -> {
            // 为每一个cid容器创建一个定时任务
            NetworkMonitorThread networkMonitorThread = new NetworkMonitorThread(cid, containerInfoMap, clusterService);
            // 放入线程池执行任务
            ScheduledFuture<?> future = scheduledExecutorService.scheduleAtFixedRate(networkMonitorThread, 0, 1, TimeUnit.SECONDS);
            // 保存任务，以方便进行定时任务的关闭
            scheduledFutureMap.put(cid, future);
        });

        return true;
    }

    @RequestMapping("/cancelNetworkMonitor")
    public boolean cancelNetworkMonitor(long clusterId, String containerName) {
        // 先获取全部的容器id
        Set<String> cids = clusterService.getContainerIdsByClusterId(clusterId, containerName);
        if(cids == null || cids.isEmpty()) {
            return false;
        }

        cids.forEach(cid -> {
            scheduledFutureMap.get(cid).cancel(true);
            scheduledFutureMap.remove(cid);
            containerInfoMap.remove(cid);
        });

        return true;
    }


    @RequestMapping(value = "/createCluster", method = RequestMethod.POST)
    public long createCluster(@RequestBody ClusterConfig clusterConfig) {
        // 创建一个解析器
        long l = clusterService.initContainers(clusterConfig);
        // 使用json进行解析
        return l;
    }

    @RequestMapping(value = "/getContainerIds")
    public Set<String> getContainerIds(long clusterId, String containerName) {
        return clusterService.getContainerIdsByClusterId(clusterId, containerName);
    }

    @RequestMapping(value = "/createNetwork")
    public String createNetwork(long clusterId, String appName) {
        // TODO: 测试用
        // 判断是否存在
        return clusterService.createNetwork(clusterId, appName);
    }

    private static class NetworkMonitorThread implements Runnable {
        private String containerId;
        private ConcurrentHashMap<String, ContainerInfo> map;
        private ClusterService clusterService;

        public NetworkMonitorThread(String containerId, ConcurrentHashMap<String, ContainerInfo> map, ClusterService clusterService) {
            this.containerId = containerId;
            this.map = map;
            this.clusterService = clusterService;
        }

        @Override
        public void run() {
            NetworkInfo networkInfo = clusterService.getNetworkInfoByContainerId(containerId);
            String cpu = clusterService.cpu(containerId);
            String mem = clusterService.mem(containerId);
            ContainerInfo containerInfo = new ContainerInfo();
            containerInfo.setCpu(cpu);
            containerInfo.setId(containerId);
            containerInfo.setNetworkInfo(networkInfo);
            containerInfo.setMem(mem);
            map.put(containerId, containerInfo);
        }

        public void setContainerId(String containerId) {
            this.containerId = containerId;
        }
    }

    @RequestMapping("/edgecenter")
    public Map<String, String> edgeCenter() {
        HashMap<String, String> edgeInfo = new HashMap<>();
        edgeInfo.put("cpu", clusterService.cpu());
        edgeInfo.put("mem", clusterService.mem());
        return edgeInfo;
    }

    @RequestMapping(value = "/createTopo", method = RequestMethod.POST)
    public NetworkTopology createTopo(long clusterId, long topologyId, String[] appNames, String topology) {
        // 二维数组作为一个字符串传递进来，需要进行一个解析
        NetworkTopology topo = new NetworkTopology();
        topo.setAppNames(appNames);
        topo.setTopologyId(111);
        topo.setTopology(parseString(topology));
        // 使用topo创建工具
        topologyService.createTopo(clusterId, topo);
        return topo;
    }

    @RequestMapping(value = "/delCluster", method = RequestMethod.GET)
    public String delCluster(long clusterId) {
        clusterService.removeContainersByClusterId(clusterId);
        return "success";
    }

    private int[][] parseString(String topology) {
        String[] lines = topology.split(";");
        int x = lines.length;
        int y = lines[0].split(",").length;

        int[][] topo = new int[x][y];
        for (int i = 0; i < x; i++) {
            String[] integers = lines[i].split(",");
            for (int j = 0; j < y; j++) {
                topo[i][j] = Integer.parseInt(integers[j]);
            }
        }
        return topo;
    }
}
